/* 
Copyright (C) 2018 Cyrus Welgan 
cyrus@c-csg.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/
#include "c-csg.h"
#include "bb.h"


struct RBOX
{
	double bb[6];
	int valid;
};

void bb_normalize(double *bbox)
{
	//keep lower values on the left and bottom
	double s;

	if(bbox[0] > bbox[3])
		{
			s = bbox[0];
			bbox[0] = bbox[3];
			bbox[3] = s;
		}

	if(bbox[1] > bbox[4])
		{
			s = bbox[1];
			bbox[1] = bbox[4];
			bbox[4] = s;
		}

	if(bbox[2] > bbox[5])
		{
			s = bbox[2];
			bbox[2] = bbox[5];
			bbox[5] = s;
		}
}

void bb_expand(double *bbox, double *v)
{
	//expand bbox to include point
	if(v[0] < bbox[0])
		{
			bbox[0] = v[0];
		}
	else if(v[0] > bbox[3])
		{
			bbox[3] = v[0];
		}

	if(v[1] < bbox[1])
		{
			bbox[1] = v[1];
		}
	else if(v[1] > bbox[4])
		{
			bbox[4] = v[1];
		}

	if(v[2] < bbox[2])
		{
			bbox[2] = v[2];
		}
	else if(v[2] > bbox[5])
		{
			bbox[5] = v[2];
		}
}

void bb_transform_bbox(double *bbox, double *sbbox, double *m)
{
	//transform 8 corner points, set up new bbox
	//012, 015, 042, 045
	//312, 315, 342, 345
	double a[3];
	double b[3];
	double c[3];
	a[0] = sbbox[0];
	a[1] = sbbox[1];
	a[2] = sbbox[2];
	m_transform(b, a, m);
	a[0] = sbbox[3];
	a[1] = sbbox[4];
	a[2] = sbbox[5];
	m_transform(c, a, m);

	if(b[0] < c[0])
		{
			bbox[0] = b[0];
			bbox[3] = c[0];
		}
	else
		{
			bbox[3] = b[0];
			bbox[0] = c[0];
		}

	if(b[1] < c[1])
		{
			bbox[1] = b[1];
			bbox[4] = c[1];
		}
	else
		{
			bbox[4] = b[1];
			bbox[1] = c[1];
		}

	if(b[2] < c[2])
		{
			bbox[2] = b[2];
			bbox[5] = c[2];
		}
	else
		{
			bbox[5] = b[2];
			bbox[2] = c[2];
		}

	a[0] = sbbox[0];
	a[1] = sbbox[4];
	a[2] = sbbox[5];
	m_transform(b, a, m);
	bb_expand(bbox, b);
	a[0] = sbbox[3];
	a[1] = sbbox[1];
	a[2] = sbbox[2];
	m_transform(b, a, m);
	bb_expand(bbox, b);
	a[0] = sbbox[3];
	a[1] = sbbox[4];
	a[2] = sbbox[2];
	m_transform(b, a, m);
	bb_expand(bbox, b);
	a[0] = sbbox[0];
	a[1] = sbbox[4];
	a[2] = sbbox[2];
	m_transform(b, a, m);
	bb_expand(bbox, b);
	a[0] = sbbox[0];
	a[1] = sbbox[1];
	a[2] = sbbox[5];
	m_transform(b, a, m);
	bb_expand(bbox, b);
	a[0] = sbbox[3];
	a[1] = sbbox[1];
	a[2] = sbbox[5];
	m_transform(b, a, m);
	bb_expand(bbox, b);
}

void bb_add(double *bbox, double *sbbox)
{
	//extend extents of bbox if necessary to contain sbbox
	bb_expand(bbox, &sbbox[0]);
	bb_expand(bbox, &sbbox[3]);
}

inline int bb_point_inside(double *bbox, double *v)
{
	return (
	           v[0] >= bbox[0] && v[0] <= bbox[3] &&
	           v[1] >= bbox[1] && v[1] <= bbox[4] &&
	           v[2] >= bbox[2] && v[2] <= bbox[5]
	       );
}






//does a enclose b?
int bb_encloses(double *a, double *b)
{
	if(a[0] <= b[0] && a[3] >= b[3] &&
	        a[1] <= b[1] && a[4] >= b[4] &&
	        a[2] <= b[2] && a[5] >= b[5])
		return 1;

	return 0;
}

//is a apart from b?
int bb_apart(double *a, double *b)
{
	if(a[0] >= b[3] || a[1] >= b[4] || a[2] >= b[5] ||
	        a[3] <= b[0] || a[4] <= b[1] || a[5] <= b[2]
	  )
		return 1;

	return 0;
}


int bb_empty(double *a)
{
	if(a[0] == a[3] || a[1] == a[4] || a[2] == a[5])
		return 1;

	return 0;
}



//expand bounding boxes to exactly match resolution grid
void bb_match_resolution(struct BBLIST **head, double resolution)
{
	if(!(*head))
		return;

	struct BBLIST *bl = *head;

	do
		{
			double r;
			double *box = bl->bb;
			r = fmod(box[0], resolution);
			box[0] = (box[0] - r) - resolution; 
			r = fmod(box[1], resolution);
			box[1] = (box[1] - r) - resolution;
			r = fmod(box[2], resolution);
			box[2] = (box[2] - r) - resolution;
			r = fmod(box[3], resolution);
			box[3] = (box[3] - r) + resolution;
			r = fmod(box[4], resolution);
			box[4] = (box[4] - r) + resolution;
			r = fmod(box[5], resolution);
			box[5] = (box[5] - r) + resolution;
		}
	while((bl = bl->next) != NULL);
}


void bb_list_add(struct BBLIST **head, double *bb)
{
	struct BBLIST *bl;
	bl = malloc(sizeof(struct BBLIST));
	memcpy(bl->bb, bb, sizeof(double) * 6);
	bl->next = NULL;
	bl->deleted = 0;

	if(!(*head))
		{
			head[0] = bl;
			return;
		}

	struct BBLIST *c = *head;

	while(c->next)
		c = c->next;

	c->next = bl;
}

int bb_list_len(struct BBLIST *item)
{
	int r = 0;

	if(!item)
		return r;

	do
		{
			r++;
		}
	while( (item = item->next) );

	return r;
}

void bb_list_remove(struct BBLIST **head, struct BBLIST *item)
{
	if(!head[0])
		return;

	struct BBLIST *last = NULL;
	struct BBLIST *bl;
	bl = *head;

	if(item == bl)
		{
			head[0] = item->next;
			free(item);
			return;
		}

	last = bl;
	bl = bl->next;

	do
		{
			if(bl == item)
				{
					last->next = item->next;
					free(item);
					return;
				}

			last = bl;
		}
	while((bl = bl->next) != NULL);
}

void bb_list_free(struct BBLIST **head)
{
	if(!head[0])
		return;

	struct BBLIST *bl = *head;
	struct BBLIST *temp;

	do
		{
			temp = bl->next;
			free(bl);
		}
	while((bl = temp) != NULL);

	head[0] = NULL;
}


//expand a with b
void bb_expand_box(double *a, double *b)
{
	if(a[0] > b[0])
		a[0] = b[0];

	if(a[1] > b[1])
		a[1] = b[1];

	if(a[2] > b[2])
		a[2] = b[2];

	if(a[3] < b[3])
		a[3] = b[3];

	if(a[4] < b[4])
		a[4] = b[4];

	if(a[5] < b[5])
		a[5] = b[5];
}


void bb_debug_print(struct BBLIST *l)
{
	while(l)
		{
			if(!l->deleted)
				printf("translate([%0.2f,%0.2f,%0.2f]) scale([%0.2f,%0.2f,%0.2f1]) cube();\n", l->bb[0], l->bb[1], l->bb[2], l->bb[3] - l->bb[0], l->bb[4] - l->bb[1], l->bb[5] - l->bb[2]);

			l = l->next;
		}
}

//split boxes with each other
int bb_split_box(struct BBLIST **head, struct BBLIST *a, struct BBLIST *b)
{
	struct BBLIST *wl = NULL;
	struct BBLIST *al, *bl;
	double bb[6];
	bb_list_add(&wl, a->bb);
	bb_list_add(&wl, b->bb);
	//now slice and dice our private list until nothing left to chop
	int chopped = 1;

	while(chopped)
		{
			chopped = 0;
			al = wl;

			do
				{
					if(al->deleted)
						continue;

					bl = wl;

					do
						{
							if(bl == al)
								continue;

							if(bl->deleted)
								continue;

							if(bb_apart(al->bb, bl->bb))
								continue;

							if(bb_encloses(al->bb, bl->bb))
								{
									bl->deleted = 1;
									break;
								}

							//al is box in need of chopping, against bl

							//see if al intersects bl on the x faces
							if(al->bb[0] < bl->bb[0] && al->bb[3] > bl->bb[0]) //intersects first x face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[3] = bl->bb[0];
									bb[0] = bl->bb[0];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[3] - al->bb[0] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}

							if(al->bb[0] < bl->bb[3] && al->bb[3] > bl->bb[3]) //intersects 2nd x face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[3] = bl->bb[3];
									bb[0] = bl->bb[3];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[3] - al->bb[0] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}

							if(al->bb[2] < bl->bb[2] && al->bb[5] > bl->bb[2]) //intersects first z face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[5] = bl->bb[2];
									bb[2] = bl->bb[2];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[5] - al->bb[2] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}

							if(al->bb[2] < bl->bb[5] && al->bb[5] > bl->bb[5]) //intersects 2nd z face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[5] = bl->bb[5];
									bb[2] = bl->bb[5];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[5] - al->bb[2] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}

							if(al->bb[1] < bl->bb[1] && al->bb[4] > bl->bb[1]) //intersects first y face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[4] = bl->bb[1];
									bb[1] = bl->bb[1];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[4] - al->bb[1] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}

							if(al->bb[1] < bl->bb[4] && al->bb[4] > bl->bb[4]) //intersects 2nd y face
								{
									memcpy(bb, al->bb, sizeof(double) * 6);
									al->bb[4] = bl->bb[4];
									bb[1] = bl->bb[4];
									bb_list_add(&wl, bb);
									chopped++;

									if(al->bb[4] - al->bb[1] <= 0.0)
										{
											al->deleted = 1;
											break;
										}
								}
						}
					while( (bl = bl->next)  && al->deleted == 0 );
				}
			while( (al = al->next) );
		}

	bl = head[0];

	while(bl->next)
		bl = bl->next;

	bl->next = wl;
	//since our list copied the original arguments, mark them deleted in original list
	a->deleted = 1;
	b->deleted = 1;

	/*
	printf("scratch list: \n");
	bb_debug_print(wl);
	printf("main list:\n");
	bb_debug_print(head[0]);
	*/

	if(wl)
		return 1;

	return 0;
}



inline double bb_volume(double *bb)
{
	return (bb[3] - bb[0]) * (bb[4] - bb[1]) * (bb[5] - bb[2]);
}


//find how much volume of a is within b
double bb_shared_volume(double *a, double *b)
{
	double t[6];
	memcpy(t, a, sizeof(double) * 6);

	//create intersection box
	if( t[0] < b[0] )
		t[0] = b[0];

	if(t[3] > b[3])
		t[3] = b[3];

	if(t[3] <= t[0])
		return 0.0;

	if( t[1] < b[1] )
		t[1] = b[1];

	if(t[4] > b[4])
		t[4] = b[4];

	if(t[4] <= t[1])
		return 0.0;

	if( t[2] < b[2] )
		t[2] = b[2];

	if(t[5] > b[5])
		t[5] = b[5];

	if(t[5] <= t[2])
		return 0.0;

	return bb_volume(t);
}



//merge boxes that overlap each other to a large degree, and then split those that remain
void bb_prepare_render_volume(struct BBLIST **head)
{
	struct BBLIST *bl;
	struct BBLIST *al;

	if(!head[0])
		return;

	printf("preparing render volume...\n");
	//printf("initial bb list length: %u\n",bb_list_len(*head));
	//bb_debug_print(*head);
	//remove bb that are enclosed within another or zero volume, merge close fits
	//use deleted flag rather than change the list under our feet
	al = head[0];

	do
		{
			if(al->deleted)
				continue;

			double a_vol = bb_volume(al->bb);

			if(a_vol == 0.0)
				{
					al->deleted = 1;
					continue;
				}

			bl = *head;

			do
				{
					if(bl == al)
						continue;

					if(bl->deleted)
						continue;

					if( bb_encloses(al->bb, bl->bb) )
						{
							bl->deleted = 1;
							continue;
						}

					double sv = bb_shared_volume(al->bb, bl->bb);
					double tv = bb_volume(bl->bb) + a_vol;

					if( sv / tv > 0.7 ) //arbitrary threshold
						{
							bb_expand_box(al->bb, bl->bb);
							bl->deleted = 1;
							continue;
						}
				}
			while( (bl = bl->next) && al->deleted == 0 );
		}
	while( (al = al->next)  );

	//now split the new set of boxes.. but never move this into the same loop as expand box
	al = head[0];

	do
		{
			if(al->deleted)
				continue;

			bl = *head;

			do
				{
					if(bl == al)
						continue;

					if(bl->deleted)
						continue;

					if( bb_encloses(al->bb, bl->bb) )
						{
							bl->deleted = 1;
							continue;
						}

					double sv = bb_shared_volume(al->bb, bl->bb);

					if(sv == 0.0)
						continue;

					bb_split_box(head, al, bl);	//could delete ourself
				}
			while( (bl = bl->next) && al->deleted == 0 );
		}
	while( (al = al->next)  );

	//printf("final bb list length: %u\n",bb_list_len(*head));
	//bb_debug_print(*head);
}