/* 
Copyright (C) 2018 Cyrus Welgan 
cyrus@c-csg.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef NOFREETYPE

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include "ttf.h"

#include <ft2build.h>
#include FT_FREETYPE_H
#include <ftoutln.h>
#include <ftadvanc.h>

#include "c-csg.h"

//26.6 fixed point
float ft_to_float_single(FT_Fixed f)
{
	return ( ((float)(f >> 6)) + ((float)(f & 0b111111)) / ((float)0b111111) );
}


void ft_to_float(float *dst, const FT_Vector *v)
{
	dst[0] = ft_to_float_single(v->x);
	dst[1] = ft_to_float_single(v->y);
}



int move_to( const FT_Vector  *to, void *user )
{
	struct TTF_CHAR *ttf = (struct TTF_CHAR *)user;
	float t[2];
	ft_to_float(t, to);
	ttf->p = realloc(ttf->p, sizeof(struct TTF_POINT) * (ttf->len + 1));
	ttf->p[ttf->len].flags = TTF_FLAG_START;
	ttf->p[ttf->len].x = (double)t[0];
	ttf->p[ttf->len].y = (double)t[1];
	ttf->p[ttf->len].cx = 0.0;
	ttf->p[ttf->len].cy = 0.0;
	ttf->len++;
	//printf("move to: %0.2f, %0.2f \n", (double)t[0], (double)t[1] );
	return 0;
}

int line_to( const FT_Vector  *to, void *user )
{
	struct TTF_CHAR *ttf = (struct TTF_CHAR *)user;
	float t[2];
	ft_to_float(t, to);
	ttf->p = realloc(ttf->p, sizeof(struct TTF_POINT) * (ttf->len + 1));
	ttf->p[ttf->len].flags = TTF_FLAG_LINE;
	ttf->p[ttf->len].x = (double)t[0];
	ttf->p[ttf->len].y = (double)t[1];
	ttf->p[ttf->len].cx = 0.0;
	ttf->p[ttf->len].cy = 0.0;
	ttf->len++;
	//printf("line to: %0.2f, %0.2f \n", (double)t[0], (double)t[1] );
	return 0;
}

int conic_to( const FT_Vector  *control, const FT_Vector *to, void *user )
{
	struct TTF_CHAR *ttf = (struct TTF_CHAR *)user;
	float c1[2], t[2];
	ft_to_float(c1, control);
	ft_to_float(t, to);
	ttf->p = realloc(ttf->p, sizeof(struct TTF_POINT) * (ttf->len + 1));
	ttf->p[ttf->len].flags = TTF_FLAG_BEZ;
	ttf->p[ttf->len].x = (double)t[0];
	ttf->p[ttf->len].y = (double)t[1];
	ttf->p[ttf->len].cx = (double)c1[0];
	ttf->p[ttf->len].cy = (double)c1[1];
	ttf->len++;
	//printf("conic to c1: %0.2f, %0.2f to: %0.2f, %0.2f \n",(double)c1[0], (double)c1[1],  (double)t[0], (double)t[1] );
	return 0;
}


int cubic_to( const FT_Vector  *control1, const FT_Vector  *control2, const FT_Vector *to, void *user )
{
	printf("freetype passed cubic bezier!\n");
	/*
	struct TTF_CHAR *ttf=(struct TTF_CHAR *)user;

	float c1[2],c2[2],t[2];

	ft_to_float(c1,control1);
	ft_to_float(c2,control2);
	ft_to_float(t,to);
	*/
	return 0;
}

char *utf8_to_int(char *s, uint32_t *cr)
{
	uint32_t c = *s++;
	uint32_t ct;

	if(!(c & 0b10000000))
		goto end_conversion;

	if(c >> 4 == 0b1111) //utf8 string
		{
			c &= 0b00000111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
		}
	else if( (c & 0b11100000) == 0b11100000)
		{
			c &= 0b00001111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
		}
	else if( (c & 0b11000000) == 0b11000000)
		{
			c &= 0b00011111;
			ct = *s++;

			if(!ct)
				goto end_conversion;

			c <<= 6;
			c |= ct & 0b00111111;
		}

end_conversion:
	*cr = c;
	return s;
}

struct TTF_CHAR *ttf_string_to_curves(char *s, char *file)
{
	FT_Library  library;
	FT_Face     face;
	FT_Outline outline;
	FT_Outline_Funcs ofs;
	double xpos = 0.0;
	struct TTF_CHAR *head = NULL ;
	struct TTF_CHAR *n;

	if(FT_Init_FreeType( &library ))
		{
			printf("error with freetype library\n");
			return NULL;
		}

	if(FT_New_Face( library, file, 0, &face ))
		{
			printf("error loading font\n");
			return NULL;
		}

	if(FT_Select_Charmap(face, FT_ENCODING_UNICODE))
		{
			printf("error setting unicode charmap\n");
		}

	while(*s)
		{
			uint32_t c;
			s = utf8_to_int(s, &c);
			printf("char: 0x%08x\n", c);
			int glyph_index = FT_Get_Char_Index( face, (FT_ULong) c);

			if(!glyph_index)
				printf("glyph not found!\n");

			if(FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_BITMAP | FT_LOAD_NO_SCALE))
				{
					printf("error getting glyph\n");
					continue;
				}

			if (face->glyph->format != ft_glyph_format_outline)
				{
					printf("font does not have vector outline\n");
					FT_Done_Face(face);
					FT_Done_FreeType(library);
					return NULL;
				}

			outline = face->glyph->outline;
			memset(&ofs, 0, sizeof(ofs));
			ofs.move_to = (FT_Outline_MoveTo_Func)&move_to;
			ofs.line_to = (FT_Outline_LineTo_Func)&line_to;
			ofs.conic_to = (FT_Outline_ConicTo_Func)&conic_to;
			ofs.cubic_to = (FT_Outline_CubicTo_Func)&cubic_to;
			n = malloc_or_die(sizeof(struct TTF_CHAR));
			memset(n, 0, sizeof(struct TTF_CHAR));
			n->x = xpos;
			//n->c=*s;
			FT_Outline_Decompose(&outline, &ofs, n);

			if(n->len == 0)
				{
					free(n);
				}
			else
				{
					//link new n into head
					if(head == NULL)
						head = n;
					else
						{
							struct TTF_CHAR *t = head;

							while(t->next)
								t = t->next;

							t->next = n;
						}
				}

			FT_Fixed advance;
			FT_Get_Advances(face, glyph_index, 1, FT_LOAD_NO_SCALE, &advance);
			xpos += ft_to_float_single(advance);
			FT_Outline_Done(library, &outline);
		}

	FT_Done_Face(face);
	FT_Done_FreeType(library);
	//mark winding order
	//(x2 - x1)(y2 + y1)
	struct TTF_CHAR *c;

	for(c = head; c; c = c->next)
		{
			struct TTF_POINT *p = NULL;
			double sum = 0.0;

			if(c->len == 0)
				continue;

			p = &c->p[0];

			for(int i = 0; i < c->len - 1; i++) //always ends on starting point so last index is not wanted
				{
					double nx, ny;

					if(c->p[i].flags & TTF_FLAG_START)
						{
							if(i != 0)
								{
									//printf("commit: %f\n",sum);
									if(p && sum >= 0.0)
										p->flags |= TTF_FLAG_POSITIVE;
								}

							p = &c->p[i];
							sum = 0.0;
						}

					if(c->p[i + 1].flags & TTF_FLAG_START)
						continue; //don't rob a value from the next poly

					nx = c->p[i + 1].x;
					ny = c->p[i + 1].y;
					//printf("TEST: %f\n", (nx-c->p[i].x) * (ny+c->p[i].y));
					sum += (nx - c->p[i].x) * (ny + c->p[i].y);
				}

			//printf("commit: %f\n",sum);

			if(p && sum >= 0.0)
				p->flags |= TTF_FLAG_POSITIVE;		//save last value
		}

	/*
	//print structure

	for(c=head;c;c=c->next)
	{
		if(c->len==0)
			continue;

		for(int i=0;i<c->len;i++)
		{

			if(c->p[i].flags&TTF_FLAG_START)
				printf("START\n");
			if(c->p[i].flags&TTF_FLAG_POSITIVE)
				printf("POSITIVE\n");


			if(c->p[i].flags&TTF_FLAG_START)
				printf("%f, %f\n",c->p[i].x,c->p[i].y);
			if(c->p[i].flags&TTF_FLAG_LINE)
				printf("%f, %f\n",c->p[i].x,c->p[i].y);
			if(c->p[i].flags&TTF_FLAG_BEZ) //don't print ctrl for now
				printf("%f, %f\n",c->p[i].x,c->p[i].y);
		}
	}
	*/
	return head;
}


#else
#include <stdlib.h>
#include <inttypes.h>
#include "ttf.h"

struct TTF_CHAR *ttf_string_to_curves(char *s, char *file)
{
	return NULL;
}

#endif